/*
 * Copyright 2017-2021 Thingstream AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef INC_GPS_H_
#define INC_GPS_H_

#include <stdint.h>
#include <stdbool.h>

#if defined(__cplusplus)
extern "C" {
#elif 0
}
#endif

/**
 * Initialize the GPS module and pins associated with the GPS. This should be
 * called at least once at application startup. It can safely be called at any
 * time to turn GPS functionality on or off. Explicitly disabling GPS using
 * this function will save power.
 * @param enable true if GPS should be enabled
 * @return the cookie to be passed to Gps_callback
 */
extern void *Gps_init(bool enable);

/**
 * Return the quality of the GPS fix according to NMEA 0183. A value of 0
 * indicates that GPS does not have a location fix yet.
 * @return the quality of the GPS fix or 0 if there is no fix
 */
extern uint8_t Gps_getQuality(void);

/**
 * Callback function that can be registered with a serial #Transport instance
 * in order to parse raw GPS data sent by the transport. Data can be sent in
 * any granularity: bytes, lines or partial lines.
 * @param cookie the value returned by Gps_init
 * @param data a pointer to the data to process
 * @param len the length of the data
 */
extern void Gps_callback(void* cookie, uint8_t* data, uint16_t len);

/**
 * Return the longitude of the last GPS position.
 * Only valid when Gps_getQuality() returns a non-zero value.
 * @return the longitude in micro-degrees.
 */
extern int32_t Gps_getLongitude(void);

/**
 * Return the latitude of the last GPS position.
 * Only valid when Gps_getQuality() returns a non-zero value.
 * @return the latitude in micro-degrees.
 */
extern int32_t Gps_getLatitude(void);

#ifdef MANUFACTURING_TEST_MODE
/**
 * For MANUFACTURING_TEST_MODE return the time field from the last GPS $GGA
 * seen. A GPS fix is not required.
 *
 * @return GPS time as a number ((hh * 1000) + (mm * 100) + ss)
 */
extern uint32_t Gps_time(void);
#endif /* MANUFACTURING_TEST_MODE */

/* Gps_wfi() is a wrapper around Platform_wfi that also dumps
 * gps log buffers, when enabled. Should be called while waiting
 * for gps data to become available.
 */
#ifdef DEBUG_ENABLE
extern void Gps_wfi(void);
#else
#include "platform.h"
#define Gps_wfi() Platform_wfi()
#endif

#if defined(__cplusplus)
}
#endif

#endif /* INC_GPS_H_ */
